import shutil
from pathlib import Path
from typing import Literal

import pandas as pd

from bionty._settings import settings
from bionty.dev._io import load_yaml, write_yaml

ROOT = Path(__file__).parent.parent / "versions"
VERSIONS_PATH = ROOT / "versions.yaml"
_CURRENT_PATH = ROOT / "._current.yaml"
_LNDB_PATH = ROOT / "._lndb.yaml"

LOCAL_PATH = settings.versionsdir / "local.yaml"


def latest_db_version(db: str) -> str:
    """Lookup the latest version of a database.

    Args:
        db: The database to look up the version for.

    Returns:
        The version of the database. Usually a date.
    """
    db = db.lower()

    if db == "ensembl":
        # For Ensembl, parse the current_README file
        lines = []

        for line in pd.read_csv(
            "https://ftp.ensembl.org/pub/README",
            chunksize=1,
            header=None,
            encoding="utf-8",
        ):
            lines.append(line.iloc[0, 0])
        return "-".join(lines[1].split(" ")[1:-1]).lower()

    else:
        raise NotImplementedError


def create_current(
    overwrite: bool = True, source: Literal["versions", "local"] = "local"
) -> None:
    """Writes the most recent version to the _current.yaml .

    Takes the 1st database defined in the source.

    Args:
        overwrite: Whether to overwrite the _current.yaml even if it exists already.
        source: The yaml source to use to create the _current.yaml . Defaults to local.
    """
    if not _CURRENT_PATH.exists() or overwrite:
        versions = (
            load_yaml(VERSIONS_PATH) if source == "versions" else load_yaml(LOCAL_PATH)
        )

        def _write_current_yaml(versions):
            _current = {}
            for name, db_versions in versions.items():
                if name == "version":
                    continue
                # this will only take the 1st db if multiple exists for the same entity
                db = next(iter(db_versions))
                versions = db_versions.get(db).get("versions")
                version = str(sorted(versions.keys(), reverse=True)[0])
                _current[name] = {db: version}
            return _current

        _current = _write_current_yaml(versions)
        write_yaml(_current, _CURRENT_PATH)


def create_local(overwrite: bool = True) -> None:
    """If local.yaml doesn't exist, copy from versions.yaml and create it.

    Args:
        overwrite: Whether to overwrite the current local.yaml .
    """
    if not LOCAL_PATH.exists() or overwrite:
        versions = load_yaml(VERSIONS_PATH)
        # convert all non string keys to strings
        local_versions = {}
        for entity, dbs in versions.items():
            local_versions[entity] = versions[entity]
            if entity == "version":
                continue
            for db_name, v in dbs.items():
                # list is needed here to avoid dict key change error
                for version in list(v["versions"]):
                    if isinstance(version, str):
                        continue
                    local_versions[entity][db_name]["versions"][
                        str(version)
                    ] = local_versions[entity][db_name]["versions"].pop(version)

        write_yaml(local_versions, LOCAL_PATH)


def update_local() -> None:
    """Update _local to add additional entries from the public versions.yaml table.

    Args:
        to_update_yaml: Dictionary of the current local.yaml .
    """
    to_update_yaml = load_yaml(LOCAL_PATH)

    versions = load_yaml(VERSIONS_PATH)

    for entity, dbs in versions.items():
        if entity == "version":
            continue
        if entity not in to_update_yaml:
            to_update_yaml[entity] = versions[entity]
        else:
            for db_name, v in dbs.items():
                if db_name not in to_update_yaml[entity]:
                    to_update_yaml[entity][db_name] = dbs[db_name]
                else:
                    for version in v["versions"]:
                        if (
                            str(version)
                            not in to_update_yaml[entity][db_name]["versions"]
                        ):
                            to_update_yaml[entity][db_name]["versions"][version] = v[
                                "versions"
                            ][version]

    write_yaml(to_update_yaml, LOCAL_PATH)


def create_lndb() -> None:
    """If no ._lndb file, write ._current to ._lndb for lndb."""
    if not _LNDB_PATH.exists():
        shutil.copy2(_CURRENT_PATH, _LNDB_PATH)



import pandas as pd

import bionty as bt


def test_clo_cellline_curation_ontology_id():
    df = pd.DataFrame(
        index=[
            "BFO:0000003",
            "http://www.ebi.ac.uk/efo/EFO_0002970",
            "http://www.ebi.ac.uk/efo/EFO_0002970",
            "BFO:0000004",
            "This cell line does not exist",
        ]
    )
    curated_df = bt.CellLine(database="clo", version="2022-03-21").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_clo_cellline_curation_name():
    df = pd.DataFrame(
        index=[
            "DA05702 cell",
            "DA05703 cell",
            "immortal astrocyte cell line cell",
            "15P-1 cell",
            "This cell line does not exist",
        ]
    )
    curated_df = bt.CellLine(id="name", database="clo", version="2022-03-21").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)



import hashlib
from pathlib import Path
from typing import Union


def verify_md5(file_path: Union[Path, str], expected_md5: str) -> bool:
    """Compares the md5 sum of an input file and an expected md5 sum.

    Args:
        file_path: Path to the file to compare.
        expected_md5: The expected md5 sum that the input file should be compared to.

    Returns:
        True if the md5 sums match, False otherwise.
    """
    file_md5 = calculate_md5(file_path)

    if file_md5 == expected_md5:
        return True
    else:
        return False


def calculate_md5(file_path: Union[Path, str]) -> str:
    """Calculates the md5 sum of a file.

    Args:
        file_path: Path to the file to calculate the md5 sum for.

    Returns:
        The md5 sum.
    """
    with open(file_path, "rb") as f:
        md5 = hashlib.md5()
        while True:
            data = f.read(8192)
            if not data:
                break
            md5.update(data)
        file_md5 = md5.hexdigest()

    return file_md5



import pandas as pd

import bionty as bt


def test_cl_celltype_curation_ontology_id():
    df = pd.DataFrame(
        index=[
            "BFO:0000002",
            "BFO:0000004",
            "UBERON:8480008",
            "UBERON:8480004",
            "This cell type does not exist",
        ]
    )

    curated_df = bt.CellType(database="cl", version="2022-08-16").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_cl_celltype_curation_name():
    df = pd.DataFrame(
        index=[
            "placental artery",
            "placental artery endothelium",
            "spatial region",
            "continuant",
            "This cell type does not exist",
        ]
    )
    curated_df = bt.CellType(id="name", database="cl", version="2022-08-16").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_ca_celltype_curation_ontology_id():
    df = pd.DataFrame(
        index=[
            "CL:0000084",
            "CL:0000542",
            "CL:0000052",
            "CL:0000320",
            "This cell type does not exist",
        ]
    )

    curated_df = bt.CellType(database="ca", version="2022-12-16").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_ca_celltype_curation_name():
    df = pd.DataFrame(
        index=[
            "placental epithelial cell",
            "response to estradiol",
            "capillary",
            "serosa of esophagus",
            "This cell type does not exist",
        ]
    )
    curated_df = bt.CellType(id="name", database="ca", version="2022-12-16").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)



from typing import Literal, Optional

import pandas as pd
from cached_property import cached_property

from bionty.entities._shared_docstrings import _doc_params, species_removed

from .._entity import Entity


@_doc_params(doc_entities=species_removed)
class Species(Entity):
    """Species.

    1. Species ontology
    Edits of terms are coordinated and reviewed on:
    https://www.ensembl.org/index.html

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        id: Optional[str] = None,
        database: Optional[Literal["ensembl"]] = None,
        version: Optional[str] = None,
    ):
        super().__init__(id=id, database=database, version=version)
        if self.database == "ensembl" and id is None:
            self._id = "name"

    @cached_property
    def df(self) -> pd.DataFrame:
        """DataFrame.

        See ingestion: https://lamin.ai/docs/bionty-assets/ingest/ensembl-species
        """
        url = f"https://ftp.ensembl.org/pub/{self._version}/species_EnsemblVertebrates.txt"  # noqa
        self._filepath = self._url_download(url)

        df = pd.read_csv(self._filepath, sep="\t", index_col=False)
        df.rename(
            columns={
                "#name": "name",
                "species": "scientific_name",
                "taxonomy_id": "taxon_id",
            },
            inplace=True,
        )
        df["name"] = df["name"].str.lower()
        df.insert(0, "id", "NCBI_" + df["taxon_id"].astype(str))

        return df.set_index(self._id)



"""Developer API.

.. autosummary::
   :toctree: .

   ontology_info
   latest_db_version
"""
from ._fix_index import (
    check_if_index_compliant,
    get_compliant_index_from_column,
    explode_aggregated_column_to_expand,
)  # noqa

from ._ontology import ontology_info
from ._handle_versions import (
    latest_db_version,
    create_current,
    update_local,
    create_local,
    create_lndb,
)



from lamin_logger import logger

from bionty._compat.bionty_0_8_1 import update_yaml_from_unversionized_to_0_1
from bionty._settings import settings
from bionty.dev._io import load_yaml

_LOCAL_PATH = settings.versionsdir / "local.yaml"


def sync_yaml_format():
    versions = load_yaml(_LOCAL_PATH)

    # Unversionized to 0.1.0
    if "version" not in versions.keys():
        logger.warning(
            "Detected unversionized yaml file. Upgrading to 0.1.0 yaml structure."
        )
        update_yaml_from_unversionized_to_0_1()



from collections import namedtuple
from typing import Iterable

from ._normalize import GENE_COLUMNS, PROTEIN_COLUMNS


def _lookup(values: Iterable[str]):
    """Look up a list of values via tab completion."""
    nt = namedtuple("feature", values)  # type: ignore
    return nt(**{i: i for i in values})


class lookup:
    """Look up a list of values via tab completion."""

    gene_id = _lookup(values=set(GENE_COLUMNS.values()))
    protein_id = _lookup(values=set(PROTEIN_COLUMNS.values()))



from pathlib import Path
from typing import List, Literal, Tuple, Union

from ._compat.update_yaml_format import sync_yaml_format
from ._settings import settings
from .dev._handle_versions import (
    create_current,
    create_lndb,
    create_local,
    update_local,
)
from .dev._io import load_yaml, write_yaml

ROOT = Path(__file__).parent.parent / "versions"
VERSIONS_PATH = ROOT / "versions.yaml"
_CURRENT_PATH = ROOT / "._current.yaml"
_LNDB_PATH = ROOT / "._lndb.yaml"

_LOCAL_PATH = settings.versionsdir / "local.yaml"


def update_defaults(
    new_defaults: Union[Tuple[str, str, str], List[Tuple[str, str, str]]],
    target: Literal["current", "lndb"] = "current",
) -> None:
    """Updates the _current.yaml file wih new user defaults.

    The _current.yaml stores the default databases and versions that Bionty accesses.
    This function overwrites the current defaults in the ._current.yaml file.

    Args:
        new_defaults: Triplets of (entity, database, version) tuples.
        target: The yaml file to update. Defaults to current
    """
    defaults = (
        load_yaml(_CURRENT_PATH) if target == "current" else load_yaml(_LNDB_PATH)
    )

    if isinstance(new_defaults, Tuple):  # type: ignore
        new_defaults = [new_defaults]  # type: ignore

    # TODO Validate whether new defaults are also available in the local.yaml
    for nd in new_defaults:
        entity = nd[0]
        new_db = nd[1]
        new_version = nd[2]

        defaults[entity] = {new_db: new_version}

    write_yaml(defaults, _CURRENT_PATH)


create_local(overwrite=False)
sync_yaml_format()
update_local()
create_current(overwrite=False)
create_lndb()



import os
import re
from collections import namedtuple
from pathlib import Path
from typing import Dict, Iterable, Literal, Optional

import bioregistry as br
import pandas as pd
from cached_property import cached_property
from lamin_logger import logger

from bionty._md5 import verify_md5

from ._ontology import Ontology
from ._settings import check_datasetdir_exists, check_dynamicdir_exists, settings
from .dev._fix_index import (
    check_if_index_compliant,
    explode_aggregated_column_to_expand,
    get_compliant_index_from_column,
)
from .dev._io import load_yaml, url_download

VERSIONS_PATH = Path(__file__).parent / "versions"


def _camel_to_snake(string: str) -> str:
    """Convert CamelCase to snake_case."""
    return re.sub(r"(?<!^)(?=[A-Z])", "_", string).lower()


class Entity:
    """Biological entity as an Entity.

    See :doc:`guide/index` for background.
    """

    def __init__(
        self,
        database: Optional[str],
        id: Optional[str] = None,
        version: Optional[str] = None,
        species: Optional[str] = None,
        filenames: Optional[Dict[str, str]] = None,
        *,
        prefix: Optional[str] = None,
    ):
        self._id = "id" if id is None else id
        # By default lookup allows auto-completion for the `name` field.
        # lookup column can be changed using `.lookup_col = `.
        self._lookup_col = "name"
        self._species = "human" if species is None else species
        self.filenames = filenames if filenames else None
        self.prefix = prefix

        if database:
            # We don't allow custom databases inside lamindb instances
            # because the lamindb standard should be used
            if os.getenv("LAMINDB_INSTANCE_LOADED") == 1:
                raise ValueError(
                    "Custom databases are not allowed inside lamindb instances."
                    "Check active databases using `bionty.display_active_versions`."
                )

            if br.normalize_prefix(database):
                database = br.normalize_prefix(database)
        self._set_attributes(database=database, version=version)

    def __repr__(self) -> str:
        representation = (
            f"Entity of {self.__class__.__name__}\n"
            f"Species: {self.species}\n"
            f"Database: {self.database}\n\n"
            f"Access ontology terms with '{self.__class__.__name__}.df'"
        )
        return representation

    @property
    def database(self) -> str:
        """Name of the database."""
        return self._database

    @cached_property
    def entity(self) -> str:
        """Name of the entity."""
        return _camel_to_snake(self.__class__.__name__)

    @property
    def species(self):
        """The `name` of `Species` Entity."""
        return self._species

    @property
    def version(self):
        """The `name` of `version` entity Entity."""
        return self._version

    @cached_property
    def ontology(self, **kwargs) -> Ontology:  # type:ignore
        """The Pronto Ontology object."""
        localpath = self._url_download(self._url)

        return Ontology(handle=localpath, **kwargs)

    @cached_property
    def df(self) -> pd.DataFrame:
        """Pandas DataFrame."""
        if not self._local_parquet_path.exists():
            try:
                self._url_download(self._url)
            finally:
                # Only verify md5 if it's actually available
                if len(self._md5) > 0:
                    if not verify_md5(self._ontology_download_path, self._md5):
                        logger.warning(
                            f"MD5 sum for {self._ontology_download_path} did not match"
                            f" {self._md5}! Redownloading..."
                        )
                        os.remove(self._ontology_download_path)
                        self._url_download(self._url)

        df = self._ontology_to_df(self.ontology)
        df.to_parquet(self._local_parquet_path)

        return (
            pd.read_parquet(self._local_parquet_path).reset_index().set_index(self._id)
        )

    @property
    def lookup_col(self) -> str:
        """The column that allows auto-completion."""
        return self._lookup_col

    @lookup_col.setter
    def lookup_col(self, column_name) -> None:
        """Set the lookup column."""
        self._lookup_col = column_name

    @cached_property
    def lookup(self) -> tuple:
        """Return an auto-complete object for the bionty id."""
        df = self.df.reset_index()
        if self._lookup_col not in df:
            raise AssertionError(f"No {self._lookup_col} column exists!")

        # uniquefy lookup keys
        df.index = self._uniquefy_duplicates(
            self._to_lookup_keys(df[self._lookup_col].values)
        )

        return self._namedtuple_from_dict(df)

    def _to_lookup_keys(self, x: list) -> list:
        """Convert a list of strings to tab-completion allowed formats."""
        lookup = [re.sub("[^0-9a-zA-Z]+", "_", str(i)) for i in x]
        for i, value in enumerate(lookup):
            if not value[0].isalpha():
                lookup[i] = f"LOOKUP_{value}"
        return lookup

    def _namedtuple_from_dict(
        self, df: pd.DataFrame, name: Optional[str] = None
    ) -> tuple:
        """Create a namedtuple from a dict to allow autocompletion."""
        if name is None:
            name = self.entity
        nt = namedtuple(name, df.index)  # type:ignore
        return nt(
            **{
                df.index[i]: row
                for i, row in enumerate(df.itertuples(name=name, index=False))
            }
        )

    def _uniquefy_duplicates(self, lst: Iterable) -> list:
        """Uniquefy duplicated values in a list."""
        df = pd.DataFrame(lst)
        duplicated = df[df[0].duplicated(keep=False)]
        df.loc[duplicated.index, 0] = (
            duplicated[0] + "__" + duplicated.groupby(0).cumcount().astype(str)
        )
        return list(df[0].values)

    def _ontology_to_df(self, ontology: Ontology):
        """Convert ontology to a DataFrame with id and name columns."""
        if self.prefix:
            return pd.DataFrame(
                [
                    (term.id, term.name)
                    for term in ontology.terms()
                    if term.id.startswith(f"{self.prefix}:")
                ],
                columns=["ontology_id", "name"],
            ).set_index(self._id)
        else:
            return pd.DataFrame(
                [(term.id, term.name) for term in ontology.terms()],
                columns=["ontology_id", "name"],
            ).set_index(self._id)

    @check_dynamicdir_exists
    def _url_download(self, url: str):
        """Download file from url to dynamicdir."""
        if not self._ontology_download_path.exists():
            logger.info(
                f"Downloading {self.entity} reference for the first time might take a"
                " while..."
            )
            url_download(url, self._ontology_download_path)

        return self._ontology_download_path

    def _ontology_localpath_from_url(self, url: str):
        """Get version from the ontology url."""
        version = url.split("/")[-2]
        filename = url.split("/")[-1]
        return settings.dynamicdir / f"{version}___{filename}"

    def _load_versions(
        self, source: Literal["versions", "local"] = "local"
    ) -> Dict[str, Dict[str, Dict]]:
        """Load all versions with string version keys."""
        YAML_PATH = (
            Path(f"{VERSIONS_PATH}/versions.yaml")
            if source == "versions"
            else Path(f"{settings.versionsdir}/local.yaml")
        )
        versions = load_yaml(YAML_PATH).get(self.__class__.__name__)

        versions_db: Dict[str, Dict[str, Dict]] = {}

        for db, vers in versions.items():
            versions_db[db] = {"versions": {}}
            for k in vers["versions"]:
                versions_db[db]["versions"][str(k)] = versions[db]["versions"][k]

        return versions_db

    @check_datasetdir_exists
    def _set_attributes(
        self, database: Optional[str], version: Optional[str] = None
    ) -> None:
        """Sets version, database and URL attributes for passed database and requested version.

        Args:
            database: The database to find the URL and version for.
            version: The requested version of the database.
        """
        current_defaults = (
            "._lndb.yaml"
            if os.getenv("LAMINDB_INSTANCE_LOADED") == 1
            else "._current.yaml"
        )

        ((current_database, current_version),) = (
            load_yaml(VERSIONS_PATH / current_defaults)
            .get(self.__class__.__name__)
            .items()
        )

        available_db_versions = self._load_versions(source="local")

        # Use the latest version if version is None.
        self._database = current_database if database is None else str(database)
        # Only the database was passed -> get the latest version from the available db versions  # noqa: E501
        if database and not version:
            self._version = next(
                iter(available_db_versions[self._database]["versions"])
            )
        else:
            self._version = current_version if version is None else str(version)
        self._url, self._md5 = (
            available_db_versions.get(self._database).get("versions").get(self._version)  # type: ignore  # noqa: E501
        )
        if self._url is None:
            raise ValueError(
                f"Database {self._database} version {self._version} is not found,"
                f" select one of the following: {available_db_versions}"
            )

        self._cloud_parquet_path = f"{self.species}_{self.database}_{self.version}_{self.__class__.__name__}_lookup.parquet"  # noqa: E501
        self._local_parquet_path = (
            settings.datasetdir / self._cloud_parquet_path
        )  # noqa: W503,E501
        self._ontology_download_path = (
            settings.dynamicdir
            / f"{self.species}_{self.database}_{self.version}_{self.__class__.__name__}___{self._url.split('/')[-1]}"  # noqa: E501 W503
        )

    def curate(
        self, df: pd.DataFrame, column: str = None, case_sensitive: bool = True
    ) -> pd.DataFrame:
        """Curate index of passed DataFrame to conform with default identifier.

        - If `column` is `None`, checks the existing index for compliance with
          the default identifier.
        - If `column` denotes an entity identifier, tries to map that identifier
          to the default identifier.

        Returns the DataFrame with the curated index and a boolean `__curated__`
        column that indicates compliance with the default identifier.
        """
        df = df.copy()
        orig_column = column
        if column is not None and column not in self.df.columns:
            column = self._id
            df.rename(columns={orig_column: column}, inplace=True)

        # uppercasing the target column before curating
        orig_column_values = None
        if not case_sensitive:
            if column in df.columns:
                orig_column_values = df[column].values
                df[column] = df[column].str.upper()
            else:
                orig_column_values = df.index.values
                df.index = df.index.str.upper()
        curated_df = self._curate(df=df, column=column).rename(
            columns={column: orig_column}
        )

        # change the original column values back
        if orig_column_values is not None:
            if "orig_index" in curated_df:
                curated_df["orig_index"] = orig_column_values
            else:
                curated_df[orig_column] = orig_column_values

        return curated_df

    def _curate(
        self, df: pd.DataFrame, column: str = None, agg_col: str = None
    ) -> pd.DataFrame:
        """Curate index of passed DataFrame to conform with default identifier."""
        df = df.copy()

        if agg_col is not None:
            # if provided a column with aggregated values, performs alias mapping
            alias_map = explode_aggregated_column_to_expand(
                self.df.reset_index(), aggregated_col=agg_col, target_col=self._id
            )[self._id]

        if column is None:
            # when column is None, use index as the input column
            index_name = df.index.name
            df["__mapped_index"] = (
                df.index if agg_col is None else df.index.map(alias_map)
            )
            df["orig_index"] = df.index
            df.index = df["__mapped_index"].fillna(df["orig_index"])
            del df["__mapped_index"]
            df.index.name = index_name
            matches = check_if_index_compliant(df.index, self.df.index)
        else:
            orig_series = df[column]
            df[column] = df[column] if agg_col is None else df[column].map(alias_map)
            df[column] = df[column].fillna(orig_series)
            new_index, matches = get_compliant_index_from_column(
                df=df,
                ref_df=self.df,
                column=column,
            )

            # keep the original index name as column name if exists
            # otherwise name it "orig_index"
            if df.index.name is None:
                df["orig_index"] = df.index
            else:
                df[df.index.name] = df.index
            df.index = new_index
            df.index.name = self._id
            df[column] = orig_series.values  # keep the original column untouched
        # annotated what complies with the default ID
        df["__curated__"] = matches
        # some stats for logging
        n_misses = len(matches) - matches.sum()
        frac_misses = round(n_misses / len(matches) * 100, 1)
        n_mapped = matches.sum()
        frac_mapped = 100 - frac_misses
        logger.success(f"{n_mapped} terms ({frac_mapped}%) are mapped.")
        logger.warning(f"{n_misses} terms ({frac_misses}%) are not mapped.")

        return df



from typing import Literal, Optional

import pandas as pd
from cached_property import cached_property

from .._entity import Entity
from .._settings import s3_bionty_assets


class CellMarker(Entity):
    """Cell markers.

    1. Cell Marker Ontology
    Edits of terms are coordinated and reviewed on:
    http://bio-bigdata.hrbmu.edu.cn/CellMarker/

    Args:
        species: `name` of `Species` entity Entity.
    """

    def __init__(
        self,
        species: str = "human",
        id: Optional[str] = None,
        database: Optional[Literal["cellmarker"]] = None,
        version: Optional[str] = None,
    ) -> None:
        super().__init__(id=id, database=database, version=version, species=species)
        if self.database == "cellmarker" and id is None:
            self._id = "name"

    @cached_property
    def df(self):
        """DataFrame.

        See ingestion: https://lamin.ai/docs/bionty-assets/ingest/cell-marker-2.0
        """
        cloudpath = s3_bionty_assets(self._cloud_parquet_path)
        df = pd.read_parquet(cloudpath.fspath)

        return df.set_index(self._id)



import pandas as pd

import bionty as bt


def test_cellmarker_cellmarker_curation_name_human():
    df = pd.DataFrame(
        index=["CCR7", "CD69", "CD8A", "CD45RA", "This protein does not exist"]
    )

    curated_df = bt.CellMarker(database="cellmarker", version="2.0").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_cellmarker_cellmarker_curation_name_mouse():
    df = pd.DataFrame(
        index=["Tcf4", "Cd36", "Cd34", "Cd45", "This protein does not exist"]
    )

    curated_df = bt.CellMarker(
        species="mouse", database="cellmarker", version="2.0"
    ).curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)



from typing import Literal, Optional

from .._entity import Entity
from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class Tissue(Entity):
    """Tissue.

    1. Uberon
    Edits of terms are coordinated and reviewed on:
    https://github.com/obophenotype/uberon

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        species: str = "all",
        id: str = "ontology_id",
        database: Optional[Literal["uberon"]] = None,
        version: Optional[str] = None,
    ) -> None:
        super().__init__(
            id=id, database=database, version=version, species=species, prefix="UBERON"
        )



"""Bionty: Data model generator for biology.

Import the package::

   import bionty as bt

Entities:

.. autosummary::
   :toctree: .

   Gene
   Protein
   Species
   CellLine
   CellType
   CellMarker
   Tissue
   Disease
   Phenotype
   Readout

The base model for every entity class is:

.. autosummary::
   :toctree: .

   Entity

Display of currently available or used versions:

.. autosummary::
   :toctree: .

    display_available_versions
    display_active_versions

Lookup of vocabulary:

.. autosummary::
   :toctree: .

   lookup

Setting default ontologies

.. autosummary::
    :toctree: .

    update_defaults

External API:

.. autosummary::
   :toctree: .

   Ontology

Dev API:

.. autosummary::
   :toctree: .

   dev

"""

__version__ = "0.10.0"  # denote release candidate for 0.1.0 with 0.1rc1

# prints warning of python versions
from lamin_logger import py_version_warning

py_version_warning("3.8", "3.10")

from . import _sync_versions
from ._sync_versions import update_defaults

# dynamic classes
from .entities._gene import Gene
from .entities._species import Species
from .entities._protein import Protein
from .entities._cellline import CellLine
from .entities._celltype import CellType
from .entities._cellmarker import CellMarker
from .entities._tissue import Tissue
from .entities._disease import Disease
from .entities._phenotype import Phenotype
from .entities._pathway import Pathway
from .entities._readout import Readout

# tools
from ._normalize import NormalizeColumns
from ._entity import Entity
from ._ontology import Ontology
from ._lookup import lookup
from ._display_versions import display_active_versions, display_available_versions

# dev
from . import dev



from bionty._settings import settings
from bionty.dev._io import load_yaml, write_yaml

_LOCAL_PATH = settings.versionsdir / "local.yaml"


def update_yaml_from_unversionized_to_0_1() -> None:
    """Update an old YAML format (unversionized) to version 0.1.

    Args:
        old_yaml: The unversionized yaml.
    """
    old_yaml = load_yaml(_LOCAL_PATH)
    new_yaml = {}

    new_yaml["version"] = "0.1.0"

    for entities, values in old_yaml.items():
        new_values = {}

        for sub_section, sub_values in values.items():
            new_sub_values = {}

            if "versions" in sub_values:
                new_versions = {}

                for version, url in sub_values["versions"].items():
                    new_versions[version] = [url, ""]

                new_sub_values["versions"] = new_versions

                for key, value in sub_values.items():
                    if key != "versions":
                        new_sub_values[key] = value
            else:
                new_sub_values = sub_values
            new_values[sub_section] = new_sub_values
        new_yaml[entities] = new_values  # type: ignore

    write_yaml(new_yaml, _LOCAL_PATH, sort_keys=False, default_flow_style=False)



import pandas as pd

import bionty as bt


def test_pw_pathway_curation_ontology_id():
    df = pd.DataFrame(
        index=[
            "Reactome:R-HSA-73780",
            "KEGG:00730",
            "GO:0006385",
            "OMIM:236130",
            "This pathway does not exist",
        ]
    )
    curated_df = bt.Pathway(database="pw", version="7.74").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)



import warnings
from pathlib import Path
from typing import BinaryIO, Optional, Union

import pronto


class Ontology(pronto.Ontology):
    """Interface with ontologies via pronto.

    Also see: https://pronto.readthedocs.io/en/stable/api/pronto.Ontology.html

    Args:
        handle: Path to an ontology file.
        import_depth: The maximum depth of imports to resolve in the ontology tree.
        timeout: The timeout in seconds to use when performing network I/O.
        threads: The number of threads to use when parsing.
        url: The url of ontology.
        prefix: Dev only -> prefix for get_term.
    """

    def __init__(
        self,
        handle: Union[str, Path, BinaryIO, None] = None,
        import_depth: int = -1,
        timeout: int = 100,
        threads: Optional[int] = None,
        prefix: str = "",
    ) -> None:
        self._prefix = prefix
        warnings.filterwarnings("ignore", category=pronto.warnings.ProntoWarning)
        super().__init__(
            handle=handle, import_depth=import_depth, timeout=timeout, threads=threads
        )

    def write_obo(self):
        """Write ontology to dynamicdir/{filename}.obo file."""
        filepath = Path(self.path)
        filename = filepath.name.replace(".owl", ".obo")
        with open(filepath.parent / filename, "wb") as f:
            self.dump(f, format="obo")

    def get_term(self, term):
        """Search an ontology by its id."""
        try:
            return super().get_term(term)
        except KeyError:
            return super().get_term(f"{self._prefix}{term.replace(':', '_')}")

    def _list_subclasses(self, term, distance=1, with_self=False):
        """Subclasses of a term."""
        termclass = self.get_term(term)
        return list(termclass.subclasses(distance=distance, with_self=with_self))



import pandas as pd

import bionty as bt


def test_hp_phenotype_curation_ontology_id():
    df = pd.DataFrame(
        index=[
            "BFO:0000001",
            "HGNC:9958",
            "HGNC:8582",
            "BFO:0000006",
            "This phenotype does not exist",
        ]
    )
    curated_df = bt.Phenotype(database="hp", version="2023-01-27").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_hp_phenotype_curation_name():
    df = pd.DataFrame(
        index=[
            "eukaryotic protein",
            "Fixated interest with abnormal intensity",
            "Hand-leading gestures",
            "Idiosyncratic language",
            "This phenotype does not exist",
        ]
    )
    curated_df = bt.Phenotype(id="name", database="hp", version="2023-01-27").curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)



from textwrap import dedent


def _doc_params(**kwds):  # pragma: no cover
    """\
    Docstrings should start with "\" in the first line for proper formatting.
    """

    def dec(obj):
        obj.__orig_doc__ = obj.__doc__
        obj.__doc__ = dedent(obj.__doc__).format_map(kwds)
        return obj

    return dec


doc_entites = """\
species: `name` of `Species` entity Entity.
        id: Field name that should constitute the primary reference for each value.
            It will also be the primary key in the corresponding SQL Entity.
        database: The key of the database in the local.yml versions file.
                  Get all available databases with `bionty.display_available_versions`.
        version: The version of the ontology. Typically a date or an actual version.
                  Get available versions with `bionty.display_available_versions`.
"""
species_removed_tmp = "\n".join(doc_entites.split("\n")[1:]).split("\n")
species_removed_tmp[0] = species_removed_tmp[0].removeprefix("        ")
species_removed = "\n".join(species_removed_tmp)



import pandas as pd

import bionty as bt


def test_ensemble_gene_curation_hgnc_id():
    data = {
        "gene symbol": ["A1CF", "A1BG", "FANCD1", "corrupted"],
        "hgnc id": ["HGNC:24086", "HGNC:5", "HGNC:1101", "corrupted"],
        "ensembl_gene_id": [
            "ENSG00000148584",
            "ENSG00000121410",
            "ENSG00000188389",
            "corrupted",
        ],
    }
    df = pd.DataFrame(data).set_index("ensembl_gene_id")
    curated_df = bt.Gene(id=bt.lookup.gene_id.hgnc_id).curate(df, column="hgnc id")

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, False])

    assert curation.equals(expected_series)



import pandas as pd

import bionty as bt


def test_ensemble_species_curation_ontology_id():
    df = pd.DataFrame(
        index=[
            "NCBI_80966",
            "NCBI_211598",
            "NCBI_241587",
            "NCBI_44394",
            "This species does not exist",
        ]
    )
    curated_df = bt.Species(id="id", database="ensembl", version="release-108").curate(
        df
    )

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)


def test_ensemble_species_curation_name():
    df = pd.DataFrame(
        index=[
            "spiny chromis",
            "silver-eye",
            "platyfish",
            "california sea lion",
            "This species does not exist",
        ]
    )
    curated_df = bt.Species(
        id="name", database="ensembl", version="release-108"
    ).curate(df)

    curation = curated_df["__curated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert curation.equals(expected_series)



import os
import urllib.request
from pathlib import Path
from typing import Dict, List
from urllib.error import HTTPError, URLError

import yaml  # type:ignore

VERSIONS_FILE_PATH = Path(f"{os.getcwd()}/bionty/versions/versions.yaml")


def get_yaml_key_values(
    yaml_file_path: Path, key: str = "versions"
) -> List[Dict[str, str]]:
    """Extracts the version pairs from the versions file.

    Args:
        yaml_file_path: Path to the yaml file.
        key: The key of all versions

    Returns:
        A list of Dictionaries of versions to URLs.
    """
    with open(yaml_file_path, "r") as stream:
        versions = yaml.safe_load(stream)
    key_values = []

    for k, v in versions.items():
        if k == key:
            key_values.append(v)
        elif isinstance(v, dict):

            def get_key_values_from_dict(dictionary: Dict, key: str):
                key_values = []
                for k, v in dictionary.items():
                    if k == key:
                        key_values.append(v)
                    elif isinstance(v, dict):
                        nested_key_values = get_key_values_from_dict(v, key)
                        key_values.extend(nested_key_values)
                return key_values

            nested_key_values = get_key_values_from_dict(v, key)
            key_values.extend(nested_key_values)

    return key_values


versions_only = get_yaml_key_values(VERSIONS_FILE_PATH.resolve(), key="versions")

failed_urls = []
for pair in versions_only:
    for url_md5s in pair.values():
        url = url_md5s[0]
        if url.startswith("http"):
            try:
                assert urllib.request.urlopen(url, timeout=1000).getcode() == 200
            except URLError:
                print(f"URL: {url} is currently not accessible.")
                pass
            except (AssertionError, ValueError, HTTPError) as e:
                failed_urls.append([url, e])

if len(failed_urls) != 0:
    for fail in failed_urls:
        print(fail)
    raise AssertionError(f"{len(failed_urls)} URLs failed.")



import pandas as pd

from bionty import Gene, Protein


def test_gene_curate():
    data = {
        "gene symbol": ["A1CF", "A1BG", "FANCD1", "corrupted"],
        "hgnc id": ["HGNC:24086", "HGNC:5", "HGNC:1101", "corrupted"],
        "ensembl_gene_id": [
            "ENSG00000148584",
            "ENSG00000121410",
            "ENSG00000188389",
            "corrupted",
        ],
    }
    df_orig = pd.DataFrame(data).set_index("ensembl_gene_id")
    df_orig.index.name = None

    curated_df = Gene().curate(df_orig)
    assert "orig_index" in curated_df.columns


def test_protein_curate():
    data = {"uniprot id": ["O00444", "O00453", "corrupted"]}
    df_orig = pd.DataFrame(data).set_index("uniprot id")

    curated_df = Protein().curate(df_orig)
    assert curated_df["__curated__"].tolist() == [True, True, False]



import pandas as pd
import pytest

from bionty.dev import (
    explode_aggregated_column_to_expand,
    get_compliant_index_from_column,
)

data = {
    "gene symbol": ["A1CF", "A1BG", "FANCD1", "corrupted"],
    "hgnc id": ["HGNC:24086", "HGNC:5", "HGNC:1101", "corrupted"],
    "ensembl_gene_id": [
        "ENSG00000148584",
        "ENSG00000121410",
        "ENSG00000188389",
        "corrupted",
    ],
}
df_orig = pd.DataFrame(data).set_index("ensembl_gene_id")

df = pd.DataFrame(
    {
        "gene symbol": ["corrupted"],
        "hgnc id": ["corrupted"],
        "ensembl_gene_id": ["corrupted1"],
    }
)


def test_get_compliant_index_from_column():
    with pytest.raises(AssertionError):
        get_compliant_index_from_column(df, ref_df=df_orig, column="x")
    with pytest.raises(AssertionError):
        get_compliant_index_from_column(df, ref_df=df_orig, column="ensembl_gene_id")


def test_explode_aggregated_column_to_expand():
    assert "index" in explode_aggregated_column_to_expand(
        df, aggregated_col="ensembl_gene_id"
    )
    df.index.name = "x"
    assert "x" in explode_aggregated_column_to_expand(
        df, aggregated_col="ensembl_gene_id"
    )



import bionty as bt


def test_ontology_info():
    info = bt.dev.ontology_info("go")
    assert info["version"] == "2022-12-04"
    assert (
        info["versionIri"]
        == "http://purl.obolibrary.org/obo/go/releases/2022-12-04/go.owl"  # noqa: W503
    )
